# combo - content management system
# Copyright (C) 2015  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

from django.utils.translation import ugettext_lazy as _
from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError

from combo.utils import requests

from .models import BaseCubesChart, ChartCell
from .utils import get_cubes, get_cube, get_drilldown


class ChartForm(forms.ModelForm):
    class Meta:
        model = ChartCell
        fields = ('title', 'url')

    def __init__(self, *args, **kwargs):
        super(ChartForm, self).__init__(*args, **kwargs)
        available_charts = []
        for site_key, site_dict in settings.KNOWN_SERVICES.get('bijoe').items():
            result = requests.get('/visualization/json/',
                    remote_service=site_dict, without_user=True,
                    headers={'accept': 'application/json'}).json()
            available_charts.extend([(x['path'], x['name']) for x in result])
        available_charts.sort(key=lambda x: x[1])
        self.fields['url'].widget = forms.Select(choices=available_charts)


class CubesBarChartForm(forms.ModelForm):
    EMPTY = [(u'', _('None'))]

    class Meta:
        model = BaseCubesChart
        fields = ('title', 'url', 'cube', 'aggregate1', 'drilldown1', 'drilldown2',
                  'other_parameters')

    def __init__(self, *args, **kwargs):
        super(CubesBarChartForm, self).__init__(*args, **kwargs)
        for field in ('cube', 'aggregate1', 'drilldown1', 'drilldown2'):
            self.fields[field] = forms.ChoiceField(
                label=self.fields[field].label,
                initial=self.fields[field].initial,
                required=False,
                choices=self.EMPTY)
        if getattr(settings, 'CUBES_URL', None):
            cube_choices = self.get_cubes_choices()
            if cube_choices:
                self.fields['cube'].choices = cube_choices
            aggregate1_choices = self.get_aggregate_choices()
            # If there is no choice, hide the selector
            if not aggregate1_choices or len(aggregate1_choices) < 3:
                self.fields['aggregate1'].widget = forms.HiddenInput()
            if aggregate1_choices:
                self.fields['aggregate1'].choices = aggregate1_choices
            drilldown_choices = self.get_drilldown_choices()
            if drilldown_choices:
                self.fields['drilldown1'].choices = drilldown_choices
                self.fields['drilldown2'].choices = drilldown_choices

    def clean(self):
        cleaned_data = self.cleaned_data
        if getattr(settings, 'CUBES_URL', None):
            aggregate1_choices = self.get_aggregate_choices()
            # If there is no choice, autoselect
            if aggregate1_choices and len(aggregate1_choices) == 2:
                cleaned_data['aggregate1'] = aggregate1_choices[1][0]
        return cleaned_data

    def clean_other_parameters(self):
        other_parameters = self.cleaned_data['other_parameters']
        if other_parameters:
            try:
                decoded = json.loads(other_parameters)
                assert isinstance(decoded, dict)
                for key, value in decoded.iteritems():
                    assert isinstance(key, unicode)
                    assert isinstance(value, unicode)
            except (ValueError, AssertionError):
                raise ValidationError(_('Other parameter must be a JSON object containing only '
                                        'strings'))
        return other_parameters

    def get_cubes_choices(self):
        cubes = get_cubes()
        return self.EMPTY + [(cube['name'], cube.get('label')) for cube in cubes]

    def get_aggregate_choices(self):
        cube = self.data.get(self.add_prefix('cube')) if self.data else self.instance.cube
        if cube:
            cube = get_cube(cube)
            if cube:
                return self.EMPTY + [(ag['name'], ag['label']) for ag in cube.get('aggregates', [])]
        return []

    def get_drilldown_choices(self):
        cube = self.data.get(self.add_prefix('cube')) if self.data else self.instance.cube
        if cube:
            choices = get_drilldown(cube)
            if choices:
                return self.EMPTY + choices
        return []
